@@ -24,7 +24,6 @@ from twisted.internet import reactor, defer
 from twisted.python import log
 
 import uuid
-import operator
 from functools import partial
 from itertools import chain
 import logging
@@ -821,6 +820,7 @@ class BytesMessageProducer(object):
         self.producing = False
         self.fh = fh
         self.offset = offset
+        self.finished = False
 
     def resumeProducing(self):
         """IPushProducer interface."""
@@ -841,7 +841,7 @@ class BytesMessageProducer(object):
         Reads a little from the file, generates a BYTES message, and pass the
         control to the reactor.  If no more data, it finishes with EOF.
         """
-        if not self.producing or self.request.cancelled:
+        if not self.producing or self.request.cancelled or self.finished:
             return
 
         if self.offset:
@@ -860,6 +860,7 @@ class BytesMessageProducer(object):
             message.type = protocol_pb2.Message.EOF
             self.request.sendMessage(message)
             self.producing = False
+            self.finished = True
 
 
 class PutContent(request.Request):
@@ -1041,12 +1042,8 @@ class Authenticate(request.Request):
         if message.type == protocol_pb2.Message.AUTH_AUTHENTICATED:
             self.done()
         elif message.type == protocol_pb2.Message.ERROR:
-            if message.error.type == protocol_pb2.Error.AUTHENTICATION_FAILED:
-                self.error(request.StorageProtocolError(
-                        "Authentication Failed"))
-            else:
-                self.error(request.StorageProtocolError(
-                        "Authentication Error:"+str(message)))
+            # as the error travels with the exception, we send all here
+            self.error(request.StorageRequestError(self, message))
         else:
             self.error(request.StorageProtocolError(
                         "Authentication Error:"+str(message)))
@@ -1150,44 +1147,49 @@ class ThrottlingStorageClient(StorageClient):
     factory = None
 
     def connectionMade(self):
-        """handle connection Made"""
-        self.factory.client = self
+        """Handle connection Made."""
+        if self.factory.client is None:
+            self.factory.client = self
         StorageClient.connectionMade(self)
 
     def connectionLost(self, reason=None):
-        """handle connection lost"""
-        self.factory.unregisterProtocol(self)
+        """Handle connection lost."""
+        if self.factory.client is self:
+            self.factory.unregisterProtocol(self)
         StorageClient.connectionLost(self, reason=reason)
 
     def write(self, data):
-        """transport API to capture bytes written"""
-        self.factory.registerWritten(len(data))
+        """Transport API to capture bytes written."""
+        if self.factory.client is self:
+            self.factory.registerWritten(len(data))
         StorageClient.write(self, data)
 
     def writeSequence(self, seq):
-        """transport API to capture bytes written in a sequence"""
-        self.factory.registerWritten(reduce(operator.add, map(len, seq)))
+        """Transport API to capture bytes written in a sequence."""
+        if self.factory.client is self:
+            self.factory.registerWritten(sum(len(x) for x in seq))
         StorageClient.writeSequence(self, seq)
 
     def dataReceived(self, data):
-        """override trnasport default to capture bytes read"""
-        self.factory.registerRead(len(data))
+        """Override transport default to capture bytes read."""
+        if self.factory.client is self:
+            self.factory.registerRead(len(data))
         StorageClient.dataReceived(self, data)
 
     def throttleReads(self):
-        """pause self.transport"""
+        """Pause self.transport."""
         self.transport.pauseProducing()
 
     def unthrottleReads(self):
-        """resume self.transport"""
+        """Resume self.transport."""
         self.transport.resumeProducing()
 
     def throttleWrites(self):
-        """pause producing"""
+        """Pause producing."""
         self.pauseProducing()
 
     def unthrottleWrites(self):
-        """resume producing"""
+        """Resume producing."""
         self.resumeProducing()
 
 
@@ -1250,7 +1252,8 @@ class ThrottlingStorageClientFactory(StorageClientFactory):
     def checkReadBandwidth(self):
         """Checks if we've passed bandwidth limits."""
         if self.readLimit is not None and \
-           self.readThisSecond > self.readLimit:
+           self.readThisSecond > self.readLimit and \
+           self.unthrottleReadsID is None:
             self.throttleReads()
             throttle_time = (float(self.readThisSecond) / self.readLimit) - 1.0
             log_debug("pause reads for: %s", str(throttle_time))
@@ -1261,7 +1264,8 @@ class ThrottlingStorageClientFactory(StorageClientFactory):
     def checkWriteBandwidth(self):
         """Checks if we've passed bandwidth limits."""
         if self.writeLimit is not None and \
-           self.writtenThisSecond > self.writeLimit:
+           self.writtenThisSecond > self.writeLimit and \
+           self.unthrottleWritesID is None:
             self.throttleWrites()
             throttle_time = (float(self.writtenThisSecond) / self.writeLimit) \
                     - 1.0
